/***************************************************************************                          
    file:           ksensorsmain.cpp
    description:    ksensors main
    begin:          09/06/2024
    Copyright:      Giuseppe Della Bianca<giusdbg@gmail.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qevent.h>
#include <qapplication.h>

#include <klocalizedstring.h>
#include <kaboutapplicationdialog.h>
#include <kwindowsystem.h>
#include <kx11extras.h>
#include <kaboutdata.h>
#include <kwindowinfo.h>

#include <netwm.h>

#include "general.h"
#include "lmsensorscfg.h"
#include "ksensorsmain.h"


#include "chart.h"


ksensorsMain::ksensorsMain(QObject *parent, const QString &name): QObject(/* parent,name */)
{
    Q_UNUSED(parent);
    Q_UNUSED(name);

    sensorsCfg = nullptr;
    sensorsWidget = nullptr;
    sensorsAlarm = nullptr;
  
    qmContext = nullptr;
    qaMinimizeRestore = nullptr;
   
  idTimer = 0;
  updateInterval = 2000;
  updateIntervalTimer = 500;

  
 sensorDGList= new sensorDevsGroups;
 sensorDGList->setMonitorized(true);
 
//    init();
}

ksensorsMain::~ksensorsMain()
{
    saveConfig ( false );

    if ( idTimer )
    {
        killTimer ( idTimer );
        idTimer = 0;
    }

 if(sensorsWidget) delete sensorsWidget;
 if(sensorsCfg) delete sensorsCfg;
 if(sensorsAlarm) delete sensorsAlarm;
 
 qmContext->deleteLater();
 qmContext = nullptr;

 if(sensorDGList) delete sensorDGList;
}

/***************************************************************************
    initialize
 ***************************************************************************/

void ksensorsMain::init()
{
    createAlarmControl();
    createWidgets();
    createMenu (qmContext, qaMinimizeRestore); 
    if(!getPar ( qsFL1( "nodockopt" ) ) ) createDockWidgets();
// createMenu();

    updateSensorsPanelsDocks( qsFL1( "_init_" ), false );

    if ( getPar( qsFL1( "charttestopt" ) ) )
    {
        testChart *wcChart = new testChart();
        wcChart->show();
    }
}

void ksensorsMain::createMenu ( QMenu *&qmMenu, QAction *&qaMinRest )
{
  const KAboutData kadAboutData = KAboutData::applicationData();
  
  qmMenu = new QMenu();

  qmMenu->addAction( QIcon::fromTheme( qsFL1( COMPICON ) ), kadAboutData.displayName() );
  qmMenu->addSeparator();  
  qmMenu->addAction( QIcon::fromTheme( qsFL1( "configure") ), i18n( "&Configure" ), this, SLOT(createConfigWidget() ));
  qmMenu->addAction( QIcon ( findIcon ( "harddisk.png" ) ), i18n( "&Save Configuration" ), this, [=, this]() { this->saveConfig( true ); } );
  qmMenu->addSeparator();
  qmMenu->addAction( QIcon::fromTheme( qsFL1( COMPICON ) ), i18n( "&About ") + kadAboutData.displayName(), this, SLOT(showAbout() ));
  qmMenu->addSeparator();
    
  qaMinRest = qmMenu->addAction(sensorsWidget ? i18n("&Close"):i18n("&Restore"), this, SLOT(minimizeRestoreWidget() ));
  qaMinRest->setEnabled( !getPar ( qsFL1( "nodockopt" ) ) );
  
  qmMenu->addAction( QIcon::fromTheme( qsFL1( "exit" ) ), i18n("&Exit"), this, SLOT(close() ));
}

void ksensorsMain::createWidgets()
{
    setGroupConf ( qsFL1( "General" ) );

// warning, noDock is tested
    if ( getEntryConf ( qsFL1( "showWidget" ), false, ksensorsSetting::AutoDefault ) )
        createShowWidget ( getEntryConf ( qsFL1( "desktop" ), -1, ksensorsSetting::AutoDefault ) );
}

void ksensorsMain::saveConfig ( bool enabSync )
{
    qDebug() << "\033[91m" << "qDebug(): ksensorsMain::saveConfig enabSync" << enabSync <<"\033[0m";  
  
    setGroupConf ( qsFL1( "General" ) );
    setEntryConf ( qsFL1( "showWidget" ), (bool) sensorsWidget );

    if (sensorsWidget)
    {
        sensorsWidget->saveGeometry();

        KWindowInfo winInfo (sensorsWidget->winId(), NET::WMDesktop);
        if ( winInfo.valid() )
            setEntryConf ( qsFL1( "desktop" ), winInfo.desktop() );
    }

    if ( enabSync )
        syncConf();
}

/***************************************************************************
    update sensors
 ***************************************************************************/

void ksensorsMain::updateSensorsStatus ( Updates updatesType, struct T_UpdateData &updateData )
{
    QString sensorDevName, sensorName;
    const QObjectList *chips;

    chips = sensorDGList->getSensorsChips();

    sensorName = sensorDevName = updateData.sensorName;

    if ( updatesType & ksensorsMain:: UpdateSensorsList)
        sensorDGList->sensorsList.clear();

    if( chips )
    {
        for ( QObjectList::ConstIterator it = chips->begin(); it != chips->end(); ++it )
        {
            SensorsList *chip = ( SensorsList * )( *it );

            if( chip )
            {              
                if ( sensorDevName.isEmpty() || ( chip->getLongName() ==  sensorDevName ) )
                {
                    sensorName.clear();

                    if ( updatesType & ksensorsMain:: UpdateSensorsList )
                        sensorDGList->sensorsList.insert ( QString::number ( chip->getClass() ) + qsFL1( "_" ) + chip->getLongName(), chip) ;

                    if ( updatesType & ksensorsMain::ReadConfig )
                    {
                        chip->readConfig();
                        chip->updateSensors( true );
                    }

                    if ( updatesType & ksensorsMain::UpdateIntervalList )
                        updateData.updateIntervalList.append ( chip->getSensorUpdateInterval( false ) );

                    if ( updatesType & ksensorsMain::GlobalTimer )
                    {
                        if ( chip->getMonitorized() )
                        {
// qDebug() << "\033[91m" << "qDebug(): ksensorsMain::updateSensorsStatus chip->getMonitorized()" << chip->getMonitorized() << " chip->getLongName() " << chip->getLongName() <<"\033[0m";

                            chip->globalTimerUpdate ( updateInterval, updateIntervalTimer );
                        }
                    }
                } // if ( sensorDevName.isEmpty() || ( chip->getLongName() ==  sensorDevName ) )


                QObjectList *list = ( QObjectList * )chip->getSensors();

                if( list )
                {
                    for ( QObjectList::ConstIterator itList = list->begin(); itList != list->end(); ++itList )
                    {
                        Sensor *sensor = ( Sensor * )( *itList );

                        if ( sensor )
                        {
                            if ( sensorName.isEmpty() || ( sensor->getLongName() ==  sensorName ) )
                            {
                                if ( updatesType & ksensorsMain::ReadConfig )
                                {
                                    Q_EMIT sensor->panelDockConfigChanged();

                                    if ( sensorsWidget )
                                        sensorsWidget->updateWidget( sensor->getLongName(), updateData.withChange );

                                    if ( sensor->getMonitorize() )
                                    {

                                    }
                                }
                            } // if ( sensorName.isEmpty() || ( sensor->getLongName() ==  sensorName ) )
                        }
                        else
                            std::abort();
                    } // for ( QObjectList::ConstIterator itList = list->begin(); itList != list->end(); ++itList )
                } // if( list )
            } // if( chip )
            else
                std::abort();;
        } // for ( QObjectList::ConstIterator it = chips->begin(); it != chips->end(); ++it )
    } // if( chips )


    if ( updatesType & ksensorsMain::ReadConfig && updateData.sensorName.isEmpty() )
    {
        if ( sensorsWidget )
            sensorsWidget->updateWidget( qsFL1( "" ), false );
    }

} // void updateSensorsStatus ( Updates updatesType, struct T_UpdateData &updateData )

void ksensorsMain::createDockWidgets()
{
  const QObjectList *chips= sensorDGList->getSensorsChips();

 countSensors= 0;
 if(chips)
 {
    for ( QObjectList::ConstIterator it = chips->begin(); it != chips->end(); ++it )       
    {
      SensorsList *chip = (SensorsList *)(*it);
       
      if ( chip )
      {
        
        QObjectList *list= (QObjectList *)chip->getSensors();
        if(list) 
        {
          for ( QObjectList::ConstIterator itList = list->begin(); itList != list->end(); ++itList )       
          {
            LMSensor *sensor = (LMSensor *)(*itList);
       
            if ( sensor )
            {
//              setGroupConf ( sensor->objectName() );
                setSensorDevGroupConf ( sensor->getShortName(), sensor->getLongName(), sensor->getSubName() );
                if ( getEntryConf ( qsFL1( "showInDock" ), false, ksensorsSetting::AutoDefault ) )
                    createDockSensor(sensor);
            }
            else
              break;
          }
        }
        
       }
       else
         break;
       
    }    
 }
    
 if(countSensors==0) createDockSensor( nullptr );
}

void ksensorsMain::createAlarmControl()
{
 sensorsAlarm= new LMSensorsAlarms(sensorDGList);
}


void ksensorsMain::minimizeRestoreWidget()
{
 if(sensorsWidget)  delete sensorsWidget;
 else               createShowWidget();
 updateMenu ( sensorsWidget );
}

void ksensorsMain::updateMenu ( bool enabMinimize )
{
  QString qsActionText;
  
  
 if(getPar ( qsFL1( "nodockopt" ) ) && !enabMinimize)
   close();
 else
 { 
    qsActionText = enabMinimize ? i18n("&Close") : i18n("&Restore");
   
    qaMinimizeRestore->setText( qsActionText );
    
    const QObjectList *list= &children();
    
    if(list)
    {
        for ( QObjectList::ConstIterator it = list->begin(); it != list->end(); ++it ) 
        {
            LMSensorDockPanel *sensor = (LMSensorDockPanel *)(*it);
 
            if ( sensor )
            {
                sensor->updateMenu( qsActionText );
            }
            else
                break;      
        }
    }
 } // else (noDock && !sensorsWidget)
}

void ksensorsMain::createShowWidget(int desktop)
{
 if(!sensorsWidget)  {
   sensorsWidget= new LMSensorsWidget(sensorDGList);
   connect( sensorsWidget, SIGNAL(rightMouseClicked(QMouseEvent *)), this, SLOT(mouseEventReceived(QMouseEvent *)) );
   connect( sensorsWidget, &LMSensorsWidget::destroyed, this, [=, this]() { this->updateMenu ( false ); } );
 }
 if ((desktop != 0) && (desktop <= KX11Extras::numberOfDesktops())) KX11Extras::setOnDesktop(sensorsWidget->winId(), desktop);
 sensorsWidget->show();
}

void ksensorsMain::createConfigWidget()
{
 if(!sensorsCfg) {
   sensorsCfg= new KSensorsCfg( this );
 }
  
 sensorsCfg->show();
}

void ksensorsMain::close()
{
//  QCoreApplication::quit();
  
  this->deleteLater();
  
//  delete this;
}

void ksensorsMain::mouseEventReceived(QMouseEvent *e)
{
 if(e->type()==QEvent::MouseButtonRelease) {
   if( e->button()==Qt::LeftButton  ) minimizeRestoreWidget();
   if( e->button()==Qt::RightButton )
   {
#if (QT_VERSION < QT_VERSION_CHECK(6, 0, 0))
       qmContext->popup(e->globalPos());
#else
       qmContext->popup(e->globalPosition().toPoint());
#endif
   }
 }
}

void ksensorsMain::createDockSensor(Sensor *sensor)
{
  if(countSensors==0) {
    const QObjectList *list= &children();
    
    if(list)
    {
      for ( QObjectList::ConstIterator it = list->begin(); it != list->end(); ++it ) 
      {
        LMSensorDockPanel *sensor = (LMSensorDockPanel *)(*it);
 
        if ( sensor )
        {
          delete sensor;        
        }
        else
          break;
        
      }
    }
  }
  const QString &name= sensor ?  sensor->getLongName() : QString();
  LMSensorDockPanel *dock= new LMSensorDockPanel( sensor, this, name );
  connect( dock, SIGNAL(mouseEvent(QMouseEvent *)), this, SLOT(mouseEventReceived(QMouseEvent *)) );
  
  dock->setParent( this );
//  dock->show();
  if ( !name.isEmpty() ) countSensors++;
}

void ksensorsMain::deleteDockSensor(LMSensorDockPanel *sensor)
{
 delete sensor;
 if(--countSensors==0) createDockSensor( nullptr );
}

void ksensorsMain::updateSensorsPanelsDocks( const QString &name, bool withChange )
{
    double seconds;

    struct T_UpdateData updateData;
    
    ksensorsMain::Updates updatesType;
    
    if( !name.isEmpty() && name != qsFL1( "General" ) && name != qsFL1( "_init_" ) )
    {

 Sensor *sensor= sensorDGList->getSensor(name);

//if((bool)sensor && sensor->getType()==LMSensor::lmTemp) {
//  setGroupConf ( name );
    if ( sensor )
        setSensorDevGroupConf ( sensor->getShortName(), sensor->getLongName(), sensor->getSubName() );       
    
    bool fShow= sensor && getEntryConf ( qsFL1( "showInDock" ), false, ksensorsSetting::AutoDefault );

   
   LMSensorDockPanel *dockSensor= findChild<LMSensorDockPanel *>( name );
   if( fShow!=(bool)dockSensor ) {
     if(dockSensor) deleteDockSensor(dockSensor);
     else           createDockSensor(sensor);
   } else {
     if(dockSensor!= nullptr )
       dockSensor->slotUpdateConfig();
   }
// }

    } // if( !name.isEmpty() && name != "General" && name != "_init_" ) 
//  else
    {
        updateData.init();
        updateData.withChange = withChange,
        updateData.sensorName = ( ( !name.isEmpty() && name != qsFL1( "General" ) && name != qsFL1( "_init_" ) ) ? name : qsFL1( "" ) );

        updatesType = ksensorsMain::ReadConfig | ksensorsMain::UpdateIntervalList;
        if( name == qsFL1( "_init_" ) )
            updatesType |= ksensorsMain::UpdateSensorsList;

        updateSensorsStatus ( updatesType, updateData );

        setGroupConf ( qsFL1( "General" ) );
        seconds = getSettingUpdateInterval ( qsFL1( "UpdateInterval" ), -1.0, true );
        updateInterval = seconds * 1000;
        updateIntervalTimer = updateInterval;

        if ( !updateData.updateIntervalList.isEmpty() )
            updateIntervalTimer = maxCommonDivisor ( updateData.updateIntervalList );


        if ( idTimer )
        {
            killTimer ( idTimer );
            idTimer = 0;
        }

        
        
        if ( !idTimer )
            idTimer = startTimer( updateIntervalTimer );

    } // else ( !name.isEmpty() )
}

/***************************************************************************
    calculate the maximum commm divisor
 ***************************************************************************/

int ksensorsMain::maxCommonDivisor ( QList<int> &intList )
{
    int intMin, intMax, intCurr, intOut;

    Q_UNUSED ( intMax );
    
    intOut = 0;
    
    intMin = *std::min_element( intList.begin(), intList.end() );
    intMax = *std::max_element( intList.begin(), intList.end() );

    for( intCurr = intMin; intCurr >= 1; intCurr-- )
    {
        if (std::all_of( intList.cbegin(), intList.cend(), [intCurr]( int intItem ){ return intItem % intCurr == 0; } ) )
        {
            intOut = intCurr;

            break;
        }
    }

    return intOut;
}

void ksensorsMain::showAbout()
{  
  KAboutApplicationDialog *dialog;
  
  const KAboutData kadAboutData = KAboutData::applicationData();

  dialog= new KAboutApplicationDialog( kadAboutData );
  dialog->exec();
  delete dialog;
}

/***************************************************************************
    get devices groups list
 ***************************************************************************/

sensorDevsGroups *ksensorsMain::getSensorDGList()
{
    return sensorDGList;
}

/***************************************************************************
    general timer event
 ***************************************************************************/

void ksensorsMain::timerEvent( QTimerEvent * )
{   
    struct T_UpdateData updateData;
  
    updateSensorsStatus ( ksensorsMain::GlobalTimer, updateData );
}
